from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AnimationOptions(HighchartsMeta):
    """An animation configuration.

    .. note::

      Animation configurations can also be defined as :class:`bool <python:bool>`, where
      ``False`` turns off animation and ``True`` defaults to a duration of 500ms and defer
      of 0ms.

    """

    def __init__(self, **kwargs):
        self._complete = None
        self._defer = None
        self._duration = None
        self._easing = None
        self._step = None

        self.complete = kwargs.get('complete', None)
        self.defer = kwargs.get('defer', None)
        self.duration = kwargs.get('duration', None)
        self.easing = kwargs.get('easing', None)
        self.step = kwargs.get('step', None)

    @property
    def complete(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to execute when the animation finishes.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._complete

    @complete.setter
    @class_sensitive(CallbackFunction)
    def complete(self, value):
        self._complete = value

    @property
    def defer(self) -> Optional[int]:
        """The number of milliseconds to defer the animation. Defaults to ``0``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if set to a negative number
        """
        return self._defer

    @defer.setter
    def defer(self, value):
        self._defer = validators.integer(value,
                                         allow_empty = True,
                                         minimum = 0,
                                         coerce_value = True)

    @property
    def duration(self) -> Optional[int]:
        """The animation duration, expressed in milliseconds. Defaults to ``500``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if set to a negative number
        """
        return self._duration

    @duration.setter
    def duration(self, value):
        self._duration = validators.integer(value,
                                            allow_empty = True,
                                            minimum = 0,
                                            coerce_value = True)

    @property
    def easing(self) -> Optional[str]:
        """The name of an easing function as defined on the JavaScript ``Math`` object.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._easing

    @easing.setter
    def easing(self, value):
        self._easing = validators.string(value, allow_empty = True)

    @property
    def step(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to execute on each step of each attribute or
        CSS property that is being animated.

        The first argument contains information about the animation and progress.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._step

    @step.setter
    @class_sensitive(CallbackFunction)
    def step(self, value):
        self._step = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'complete': as_dict.get('complete', None),
            'defer': as_dict.get('defer', None),
            'duration': as_dict.get('duration', None),
            'easing': as_dict.get('easing', None),
            'step': as_dict.get('step', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'complete': self.complete,
            'defer': self.defer,
            'duration': self.duration,
            'easing': self.easing,
            'step': self.step
        }

        return untrimmed



from typing import Optional

from validator_collection import checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.data.treemap import TreemapData
from highcharts_core.utility_classes.markers import Marker


class SunburstData(TreemapData):
    """Data point that can features a ``parent``, a ``value``, and can be sliced."""

    def __init__(self, **kwargs):
        self._marker = None
        self._sliced = None

        self.marker = kwargs.get('marker', None)
        self.sliced = kwargs.get('sliced', None)

        super().__init__(**kwargs)

    @property
    def marker(self) -> Optional[Marker]:
        """Options for the point markers of line-like series.

        :rtype: :class:`Marker` or :obj:`None <python:None>`
        """
        return self._marker

    @marker.setter
    @class_sensitive(Marker)
    def marker(self, value):
        self._marker = value

    @property
    def sliced(self) -> Optional[bool]:
        """If ``True``, displays the data point as a slice offset from the center.
        Defaults to :obj:`None <python:None>`, which behaves as ``False``.

        .. note::

          If a data point is offste, its children are also offset.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._sliced

    @sliced.setter
    def sliced(self, value):
        if value is None:
            self._sliced = None
        else:
            self._sliced = bool(value)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'SunburstData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Sunburst Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')

            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank', None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'color_value': as_dict.get('colorValue', None),
            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'parent': as_dict.get('parent', None),
            'value': as_dict.get('value', None),

            'marker': as_dict.get('marker', None),
            'sliced': as_dict.get('sliced', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'marker': self.marker,
            'sliced': self.sliced,

            'colorValue': self.color_value,
            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,
            'parent': self.parent,
            'value': self.value,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.navigation.bindings import Binding as cls
from highcharts_core.options.navigation.bindings import Bindings as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'class_name': 'CircleAnnotationBinding',
      'init': 'some-event-marker',
      'start': 'some-event-marker',
      'steps': [
          'some-event-marker',
          'some-event-marker',
          'some-event-marker'
      ],
      'end': 'some-event-marker'
    }, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Binding__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Binding__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Binding_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Binding_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('navigation/bindings/01.js', False, None),

    ('navigation/bindings/error-01.js', False, (errors.HighchartsValueError,
                                                errors.HighchartsParseError,
                                                JSONDecodeError,
                                                TypeError,
                                                ValueError)),

    ('navigation/bindings/01.js', True, None),

    ('navigation/bindings/error-01.js', True, (errors.HighchartsValueError,
                                               errors.HighchartsParseError,
                                               JSONDecodeError,
                                               TypeError,
                                               ValueError)),
])
def test_Binding_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'circle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'ellipse_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'label_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': None,
        'end': 'some-event-marker'
      },
      'rectangle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      }
    }, None),

    ({
      'circle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'ellipse_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'label_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': None,
        'end': 'some-event-marker'
      },
      'rectangle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      }
    }, ValueError),
    ({
      'circle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'ellipse_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'label_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': None,
        'end': 'some-event-marker'
      },
      'rectangle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      }
    }, ValueError),
    ({
      'circle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'ellipse_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'label_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': ['some-event-marker'],
        'end': 'some-event-marker'
      },
      'rectangle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      }
    }, ValueError),
    ({
      'circle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'ellipse_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      },
      'label_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': None,
        'end': 'some-event-marker'
      },
      'rectangle_annotation': {
        'className': None,
        'init': 'some-event-marker',
        'start': 'some-event-marker',
        'steps': [
            'some-event-marker',
            'some-event-marker'
        ],
        'end': 'some-event-marker'
      }
    }, ValueError),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_Bindings__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_Bindings__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_Bindings_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_Bindings_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('navigation/bindings/02.js', False, None),

    ('navigation/bindings/error-03.js', False, (errors.HighchartsValueError,
                                                errors.HighchartsParseError,
                                                JSONDecodeError,
                                                TypeError,
                                                ValueError)),

    ('navigation/bindings/02.js', True, None),

    ('navigation/bindings/error-03.js', True, (errors.HighchartsValueError,
                                               errors.HighchartsParseError,
                                               JSONDecodeError,
                                               TypeError,
                                               ValueError)),
])
def test_Bindings_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.no_data import NoData as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
        'attr': {
            'stroke': '#ff0000',
            'stroke-width': 2,
            'rotation': 45,
            'd': ['M', 10, 10, 'L', 30, 30, 'z']
        },
        'position': {
            'align': 'center',
            'verticalAlign': 'top',
            'x': 10,
            'y': 15
        },
        'use_html': False
    }, None),
    
    ({
          'attr': {
              'stroke': '#ff0000',
              'stroke-width': 2,
              'rotation': 45,
              'd': ['M', 10, 10, 'L', 30, 30, 'z'],
              'not a valid attribute key': 123
          },
          'position': {
              'align': 'center',
              'verticalAlign': 'top',
              'x': 10,
              'y': 15
          },
          'use_html': False
    }, (errors.HighchartsValueError,
        errors.HighchartsParseError,
        JSONDecodeError,
        ValueError)),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('no_data/01.js', False, None),
    ('no_data/02.js', False, None),

    ('no_data/error-01.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    ValueError)),
    ('no_data/error-02.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    ValueError)),

    ('no_data/01.js', True, None),
    ('no_data/02.js', True, None),

    ('no_data/error-01.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   ValueError)),
    ('no_data/error-02.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Unit tests for ``highcharts.js_literal_functions``."""

import pytest

from highcharts_core import js_literal_functions as js
from highcharts_core import constants, errors

from decimal import Decimal

import esprima
from validator_collection import checkers


@pytest.mark.parametrize('item, expected_type, expected_value, error', [
    (123, int, 123, None),
    ('this-is-a-string', str, 'this-is-a-string', None),
    (123.45, float, 123.45, None),
    (Decimal(123.45), float, 123.45, None),
    (True, bool, True, None),
    (False, bool, False, None),
    ([1, 2, 3], list, [1, 2, 3], None),
    ('null', constants.EnforcedNullType, constants.EnforcedNull, None),
    (constants.EnforcedNull, constants.EnforcedNullType, constants.EnforcedNull, None),
    (None, type(None), None, None),
    ({'test': 123}, str, str({'test': 123}), None),
])
def test_serialize_to_js_literal(item, expected_type, expected_value, error):
    if not error:
        result = js.serialize_to_js_literal(item)

        assert isinstance(result, expected_type)
        assert result == expected_value
    else:
        with pytest.raises(error):
            result = js.serialize_to_js_literal(item)


@pytest.mark.parametrize('as_str, expected, error', [
    ("""function testFunction() { return True; };""", True, None),
    ("""function() { return True; }""", True, None),
    ("""const testFunction = function() { return True; }""", True, None),
    ("""const testFunction = () => { return True; }""", True, None),
    ("""const testFunction = new Function('return True;')""", True, None),

    ("""class TestClass { constructor() { return true; }};""", True, None),
    ("""let a = 123;""", False, None),
])
def test_is_js_function_or_class(as_str, expected, error):
    if not error:
        is_function = js.is_js_function_or_class(as_str)

        assert is_function is expected
    else:
        with pytest.raises(error):
            is_function = js.is_js_function_or_class(as_str)


@pytest.mark.parametrize('as_str, expected, error', [
    ("""function testFunction() { return True; };""", True, None),
    ("""function() { return True; }""", True, None),
    ("""const testFunction = function() { return True; }""", False, None),
    ("""const testFunction = () => { return True; }""", False, None),
    ("""const testFunction = new Function('return True;')""", False, None),
    ("""let a = 123;""", False, None),
])
def test_attempt_variable_declaration(as_str, expected, error):
    if not error:
        is_function = js.attempt_variable_declaration(as_str)

        assert is_function is expected
    else:
        with pytest.raises(error):
            is_function = js.attempt_variable_declaration(as_str)


@pytest.mark.parametrize('as_dict, expected, error', [
    (None, None, None),
    ({'item1': 123}, """{\n  item1: 123\n}""", None),
    ({'item1': 123, 'item2': 456}, """{\n  item1: 123,\n  item2: 456\n}""", None),
    ({'item1': True}, """{\n  item1: true\n}""", None),
    ({'item1': False}, """{\n  item1: false\n}""", None),
    ({'item1': constants.EnforcedNull}, """{\n  item1: null\n}""", None),
    ({'item1': 'test string'}, """{\n  item1: 'test string'\n}""", None),
    ({'item1': 'function() { return True; }'},
     """{\n  item1: function() { return True; }\n}""", None),
    ({'item1': 123, 'item2': None}, """{\n  item1: 123\n}""", None),
    ({}, None, None),
    ('not-a-dict', None, TypeError),
])
def test_assemble_js_literal(as_dict, expected, error):
    if not error:
        result = js.assemble_js_literal(as_dict)
        assert result == expected
    else:
        with pytest.raises(error):
            result = js.assemble_js_literal(as_dict)


@pytest.mark.parametrize('original_str, override, expected, error', [
    ("""const testObject = {item1:123,item2:456}""", None, [('item1', 123),('item2',456)], None),
    ("""const testObj = {item1:true}""", None, [('item1',True)], None),
    ("""const testObj = {item1:undefined}""", None, [('item1',None)], None),
    ("""const testObj = {item1:null}""", None, [('item1',constants.EnforcedNull)], None),
    ("""const testObj = {item1:[1, 2, 3]}""", None, [('item1', [1, 2, 3])], None),
    ("""const testObj = {item1:{subitem:123}}""", None, [('item1',{'subitem':123})], None),
])
def test_get_key_value_pairs(original_str, override, expected, error):
    if not override:
        parsed = esprima.parseScript(original_str)
        properties = parsed.body[0].declarations[0].init.properties
    else:
        properties = override

    if not error:
        result = js.get_key_value_pairs(properties, original_str)
        assert result is not None
        assert isinstance(result, list)
        assert len(result) == len(properties)
        counter = 0
        for item in result:
            assert isinstance(item, tuple)
            assert len(item) == 2
            assert item[0] == expected[counter][0]
            assert item[1] == expected[counter][1]
            if isinstance(item[1], dict):
                assert checkers.are_dicts_equivalent(item[1],
                                                     expected[counter][1]) is True
            counter += 1
        assert result == expected
    else:
        with pytest.raises(error):
            result = js.get_key_value_pairs(properties, original_str)


@pytest.mark.parametrize('original_str, override, expected, error', [
    ("""const testObj = {item1:true}""", None, True, None),
    ("""const testObj = {item1:false}""", None, False, None),
    ("""const testObj = {item1:123}""", None, 123, None),
    ("""const testObj = {item1:'test string'}""", None, 'test string', None),
    ("""const testObj = {item1:undefined}""", None, None, None),
    ("""const testObj = {item1:null}""", None, constants.EnforcedNull, None),

    ('', 'not-a-literal', None, errors.HighchartsParseError),
])
def test_convert_js_literal_to_python(original_str, override, expected, error):
    if not override:
        parsed = esprima.parseScript(original_str)
        literal = parsed.body[0].declarations[0].init.properties[0].value
    else:
        literal = override

    if not error:
        result = js.convert_js_literal_to_python(literal, original_str)
        if expected is not None:
            assert result is not None
            assert result == expected
        else:
            assert result is None
    else:
        with pytest.raises(error):
            result = js.convert_js_literal_to_python(literal, original_str)


@pytest.mark.parametrize('original_str, override, expected, error', [
    ("""const testObj = {item1:true}""", None, True, None),
    ("""const testObj = {item1:false}""", None, False, None),
    ("""const testObj = {item1:123}""", None, 123, None),
    ("""const testObj = {item1:'test string'}""", None, 'test string', None),
    ("""const testObj = {item1:undefined}""", None, None, None),
    ("""const testObj = {item1:null}""", None, constants.EnforcedNull, None),
    ("""const testObj = {item1:[1,2]}""", None, [1,2], None),
    ("""const testObj = {item1:{subitem:'test'}}""", None, {'subitem': 'test'}, None),

    ('', 'not-a-property', None, errors.HighchartsParseError),
])
def test_convert_js_property_to_python(original_str, override, expected, error):
    if not override:
        parsed = esprima.parseScript(original_str)
        item = parsed.body[0].declarations[0].init.properties[0]
    else:
        item = override

    if not error:
        result = js.convert_js_property_to_python(item, original_str)
        if expected is not None:
            assert result is not None
            assert result == expected
        else:
            assert result is None
    else:
        with pytest.raises(error):
            result = js.convert_js_property_to_python(item, original_str)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.patterns import Pattern as cls
from highcharts_core.utility_classes.patterns import PatternOptions as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'animation': {
          'defer': 5
      },
      'pattern_options': None,
      'pattern_index': 1
    }, None),
    ({
      'animation': {
          'defer': 5
      },
      'pattern_options': {
          'aspectRatio': 0.5,
          'backgroundColor': '#999999',
          'id': 'some_id_goes_here',
          'opacity': 0.5,
          'width': 120,
          'x': 5,
          'y': 10
      },
      'pattern_index': 2
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Pattern__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_Pattern__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Pattern_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_Pattern_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/patterns/01.js', False, None),

    ('utility_classes/patterns/error-01.js', False, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),

    ('utility_classes/patterns/01.js', True, None),

    ('utility_classes/patterns/error-01.js', True, (errors.HighchartsValueError,
                                                    errors.HighchartsParseError,
                                                    JSONDecodeError,
                                                    TypeError,
                                                    ValueError)),
])
def test_Pattern_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'aspect_ratio': 0.5,
      'background_color': '#999999',
      'id': 'some_id_goes_here',
      'opacity': 0.5,
      'width': 120,
      'x': 5,
      'y': 10
     }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_PatternOptions__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_PatternOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_PatternOptions_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_PatternOptions_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/patterns/02.js', False, None),

    ('utility_classes/patterns/error-03.js', False, (errors.HighchartsValueError,
                                                     errors.HighchartsParseError,
                                                     JSONDecodeError,
                                                     TypeError,
                                                     ValueError)),

    ('utility_classes/patterns/02.js', True, None),

    ('utility_classes/patterns/error-03.js', True, (errors.HighchartsValueError,
                                                    errors.HighchartsParseError,
                                                    JSONDecodeError,
                                                    TypeError,
                                                    ValueError)),
])
def test_PatternOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



"""Tests for ``highcharts.time``."""

import pytest

from validator_collection import checkers

from highcharts_core.options.time import Time as cls
from highcharts_core.utility_classes.javascript_functions import CallbackFunction, \
    JavaScriptClass
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'timezone': 'Europe/Oslo',
      'use_utc': False
    }, None),
    ({
      'Date': JavaScriptClass(class_name = 'Date',
                              methods = ["""constructor() { return true; }"""])
    }, None),
    ({
      'Date': """class DateClass { constructor() { return true; }}"""
    }, None)
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('time/01.js', False, None),
    ('time/02.js', False, None),

    ('time/error-01.js', False, (TypeError, ValueError)),
    ('time/error-02.js', False, errors.HighchartsParseError),

    ('time/01.js', True, None),
    ('time/02.js', True, None),

    ('time/error-01.js', True, (TypeError, ValueError)),
    ('time/error-02.js', True, errors.HighchartsParseError),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)




class HighchartsError(ValueError):
    """Basic error that all other Highcharts for Python exceptions inherit from."""
    pass


class HighchartsDependencyError(ImportError):
    """:exc:`ImportError <python:ImportError>` encountered when attempting to use a
    **Highcharts for Python** method that relies on a third-party library (e.g.
    `pandas <https://pandas.pydata.org>`_,
    `PySpark <https://spark.apache.org/docs/latest/api/python/>`_, etc.) which is not
    available in the runtime environment."""
    pass


class HighchartsReadOnlyError(HighchartsError, AttributeError):
    """:exc:`AttributeError <python:AttributeError>` encountered when attempting to set
    a **Highcharts for Python** property that is only available as a read-only property.
    """


class HighchartsImplementationError(HighchartsError):
    """Error that indicates you have implemented something incorrectly in your code."""
    pass


class HighchartsValueError(HighchartsImplementationError):
    """:exc:`ValueError <python:ValueError>` encountered in the operation of Highcharts
    for Python. Typically an implementation error.
    """
    pass


class HighchartsNotSupportedError(HighchartsError, TypeError):
    """:exc:`TypeError <python:TypeError>` encountered when attempting functionality that
    is not (and is not intended to be) supported."""
    pass


class HighchartsJavaScriptError(HighchartsValueError):
    """:exc:`ValueError <python:ValueError>` encountered when a Python representation of
    some JavaScript code has been badly constructed. Typically an implementation error in
    your source code."""
    pass


class HighchartsParseError(HighchartsError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python is
    unable to parse a JavaScript object literal correctly."""
    pass


class HighchartsCSVDeserializationError(HighchartsError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python is
    unable to properly deserialize CSV data."""
    pass


class HighchartsPandasDeserializationError(HighchartsError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python is
    unable to properly deserialize Pandas data."""
    pass


class HighchartsPySparkDeserializationError(HighchartsError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python is
    unable to properly deserialize PySpark data."""
    pass


class HighchartsMissingKeyError(HighchartsParseError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python
    encounters a missing key when parsing a JavaScript object literal."""
    pass


class HighchartsCollectionError(HighchartsParseError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python
    encounters a collection with more members than expected when parsing a JavaScript
    object literal."""
    pass


class HighchartsVariableDeclarationError(HighchartsParseError):
    """:exc:`ValueError <python:ValueError>` encountered when Highcharts for Python
    tries to parse JavaScript code which is not represented as a proper variable
    declaration, and is unable to coerce the JavaScript code into a properly-formed
    variable declaration."""
    pass


class HighchartsMissingClassNameError(HighchartsJavaScriptError):
    """:exc:`ValueError <python:ValueError>` encountered when trying to serialize a
    JavaScriptClass instance to JavaScript, but the instance has no class_name provided.
    """
    pass


class HighchartsExportServerError(HighchartsValueError):
    """:exc:`ValueError <python:ValueError>` encountered when trying to use the Node
    Export Server."""
    pass


class HighchartsUnsupportedProtocolError(HighchartsExportServerError):
    """:exc:`ValueError <pythoN:ValueError>` encountered when trying to use an unsupported
    protocol to communicate with the :term:`Export Server`."""
    pass


class HighchartsUnsupportedExportError(HighchartsExportServerError):
    """:exc:`ValueError <python:ValueError>` encountered when trying to export a series type that 
    is not yet supported by the :term:`Export Server`.
    """
    pass

class HighchartsUnsupportedExportTypeError(HighchartsExportServerError):
    """:exc:`ValueError <python:ValueError>` encountered when requesting an unsupported
    image type from a :term:`Export Server`."""
    pass


class HighchartsUnsupportedConstructorError(HighchartsExportServerError):
    """:exc:`ValueError <python:ValueError>` encountered when supplying an unsupported
    constructor to an :term:`Export Server` instance."""
    pass


class HighchartsMissingExportSettingsError(HighchartsExportServerError):
    """:exc:`ValueError <python:ValueError>` encountered when attempting to
    programmatically export a chart image, but key settings were not supplied to the
    :class:`ExportServer` instance."""
    pass


class HighchartsMissingSeriesError(HighchartsValueError):
    """:exc:`ValueError <python:ValueError>` encountered when trying to reference a
    series that does not actually exist in the chart."""
    pass


from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.options.plot_options.link import LinkOptions
from highcharts_core.utility_classes.zones import Zone
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.utility_classes.events import SimulationEvents


class LayoutAlgorithm(HighchartsMeta):
    """Configuration of how to lay out the Network Graph."""

    def __init__(self, **kwargs):
        self._approximation = None
        self._attractive_force = None
        self._enable_simulation = None
        self._friction = None
        self._gravitational_constant = None
        self._initial_position_radius = None
        self._initial_positions = None
        self._integration = None
        self._link_length = None
        self._max_iterations = None
        self._max_speed = None
        self._repulsive_force = None
        self._theta = None
        self._type = None

        self.approximation = kwargs.get('approximation', None)
        self.attractive_force = kwargs.get('attractive_force', None)
        self.enable_simulation = kwargs.get('enable_simulation', None)
        self.friction = kwargs.get('friction', None)
        self.gravitational_constant = kwargs.get('gravitational_constant', None)
        self.initial_position_radius = kwargs.get('initial_position_radius', None)
        self.initial_positions = kwargs.get('initial_positions', None)
        self.integration = kwargs.get('integration', None)
        self.link_length = kwargs.get('link_length', None)
        self.max_iterations = kwargs.get('max_iterations', None)
        self.max_speed = kwargs.get('max_speed', None)
        self.repulsive_force = kwargs.get('repulsive_force', None)
        self.theta = kwargs.get('theta', None)
        self.type = kwargs.get('type', None)

    @property
    def approximation(self) -> Optional[str]:
        """Approximation used to calculate repulsive forces affecting nodes.

        When :obj:`None <python:None>`, when calculateing net force, nodes are compared
        against each other, which gives ``O(N^2)`` complexity. Using ``barnes-hut``
        approximation, we decrease this to ``O(N log N)``, but the resulting graph will
        have a different layout.

        .. note::

          Barnes-Hut approximation divides space into rectangles via quad tree, where
          forces exerted on nodes are calculated directly for nearby cells, and for all
          others, cells are treated as a separate node with center of mass.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._approximation

    @approximation.setter
    def approximation(self, value):
        self._approximation = validators.string(value, allow_empty = True)

    @property
    def attractive_force(self) -> Optional[CallbackFunction]:
        """JavaScript function which calculates the attraction force applied on a node
        which is conected to another node by a link.

        The (JavaScript) function should be passed two arguments:

          * ``d`` - which is the current distance between two nodes
          * ``k`` - which is the desired distance between two nodes

        If :obj:`None <python:None>`, defaults to:

          .. code-block:: javascript

            function (d, k) {
                return k * k / d;
            }

        If :meth:`LayoutAlgorithm.integration` is ``'verlet'``, then if
        :obj:`None <python:None>` defaults to:

          .. code-block:: javascript

            function (d, k) {
                return (k - d) / d;
            }

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._attractive_force

    @attractive_force.setter
    @class_sensitive(CallbackFunction)
    def attractive_force(self, value):
        self._attractive_force = value

    @property
    def enable_simulation(self) -> Optional[bool]:
        """If ``True``, enables live simulation of the algorithm's implementation. All
        nodes are animated as the force applies to them. Defaults to ``False``.

        .. warning::

          EXPERIMENTAL!

        :rtype: :class:`bool <python:bool>`
        """
        return self._enable_simulation

    @enable_simulation.setter
    def enable_simulation(self, value):
        if value is None:
            self._enable_simulation = None
        else:
            self._enable_simulation = bool(value)

    @property
    def friction(self) -> Optional[int | float | Decimal]:
        """Friction applied on forces to prevent nodes rushing to fast to the desired
        positions. Defaults to ``-0.981``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._friction

    @friction.setter
    def friction(self, value):
        self._friction = validators.numeric(value, allow_empty = True)

    @property
    def gravitational_constant(self) -> Optional[int | float | Decimal]:
        """Gravitational const used in the barycenter force of the algorithm. Defaults to
        ``0.0625``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._gravitational_constant

    @gravitational_constant.setter
    def gravitational_constant(self, value):
        self._gravitational_constant = validators.numeric(value, allow_empty = True)

    @property
    def initial_position_radius(self) -> Optional[int | float | Decimal]:
        """When :meth:`LayoutAlgorithm.initial_positions` is set to ``'circle'``, this
        setting is the distance from the center of the circle at which nodes will be
        created. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._initial_position_radius

    @initial_position_radius.setter
    def initial_position_radius(self, value):
        self._initial_position_radius = validators.numeric(value, allow_empty = True)

    @property
    def initial_positions(self) -> Optional[str]:
        """Initial layout algorithm for positioning nodes. Defaults to ``'circle'``.

        Accepts the following options:

          * ``"circle"``
          * ``"random"``
          * a JavaScript function where positions should be set on each node
            (``this.nodes``) as ``node.plotX`` and ``node.plotY``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._initial_positions

    @initial_positions.setter
    def initial_positions(self, value):
        self._initial_positions = validators.string(value, allow_empty = True)

    @property
    def integration(self) -> Optional[str]:
        """Integration type. Defaults to ``'euler'``.

        Available options are:

          * ``'euler'``
          * ``'verlet'``

        Integration determines how forces are applied on particles. In Euler integration,
        force is applied directly as ``newPosition += velocity``;. In Verlet integration,
        new position is based on the previous posittion without velocity:
        ``newPosition += previousPosition - newPosition``.

        Note that different integrations give different results as forces are different.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._integration

    @integration.setter
    def integration(self, value):
        if not value:
            self._integration = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['euler', 'verlet']:
                raise errors.HighchartsValueError(f'integration expects either "euler" '
                                                  f'or "verlet". Was: {value}')
            self._integration = value

    @property
    def link_length(self) -> Optional[int | float | Decimal]:
        """Ideal length (px) of the link between two nodes. When
        :obj:`None <python:None>`, length is calculated (in JavaScript) as:
        ``Math.pow(availableWidth * availableHeight / nodesLength, 0.4);``

        .. note::

          Because of the algorithm specification, length of each link might be not exactly
          as specified.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._link_length

    @link_length.setter
    def link_length(self, value):
        self._link_length = validators.numeric(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def max_iterations(self) -> Optional[int]:
        """Maximum number of iterations before algorithm will stop. In general, the
        algorithm should find positions sooner, but when rendering huge number of nodes,
        it is recommended to increase this value as finding perfect graph positions can
        require more time.

        Defaults to ``1000``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._max_iterations

    @max_iterations.setter
    def max_iterations(self, value):
        self._max_iterations = validators.integer(value,
                                                  allow_empty = True,
                                                  minimum = 1)

    @property
    def max_speed(self) -> Optional[int | float | Decimal]:
        """Maximum speed that a node can attain in one iteration. Defaults to ``10``.

        In terms of simulation, it's a maximum translation (in pixels) that a node can
        move (in both x and y dimensions). While friction is applied on all nodes,
        ``max_speed`` is applied only for nodes that move very fast, for example, small or
        disconnected ones.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_speed

    @max_speed.setter
    def max_speed(self, value):
        self._max_speed = validators.numeric(value,
                                             allow_empty = True,
                                             minimum = 0)

    @property
    def repulsive_force(self) -> Optional[CallbackFunction]:
        """JavaScript function which calculates the repulsive force applied on a node
        which is conected to another node by a link.

        The (JavaScript) function should be passed two arguments:

          * ``d`` - which is the current distance between two nodes
          * ``k`` - which is the desired distance between two nodes

        If :obj:`None <python:None>`, defaults to:

          .. code-block:: javascript

            function (d, k) {
                return k * k / d;
            }

        If :meth:`LayoutAlgorithm.integration` is ``'verlet'``, then if
        :obj:`None <python:None>` defaults to:

          .. code-block:: javascript

            function (d, k) {
                return (k - d) / d * (k > d ? 1 : 0)
            }

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._repulsive_force

    @repulsive_force.setter
    @class_sensitive(CallbackFunction)
    def repulsive_force(self, value):
        self._repulsive_force = value

    @property
    def theta(self) -> Optional[int | float | Decimal]:
        """Deteremines when distance between cell and node is small enough to caculate
        forces. Defaults to ``0.5``.

        The value of theta is compared directly with quotient ``s / d``, where ``s`` is
        the size of the cell, and ``d`` is the distance between the center of the cell's
        mass and the currently compared node.

        .. warning::

          Applies only to the ``barnes-hut`` :meth:`LayoutAlgorithm.approximation`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._theta

    @theta.setter
    def theta(self, value):
        self._theta = validators.numeric(value, allow_empty = True)

    @property
    def type(self) -> Optional[str]:
        """Type of algorithm used when positioning nodes. Defaults to
        ``'reingold-fruchterman'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        self._type = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'approximation': as_dict.get('approximation', None),
            'attractive_force': as_dict.get('attractiveForce', None),
            'enable_simulation': as_dict.get('enableSimulation', None),
            'friction': as_dict.get('friction', None),
            'gravitational_constant': as_dict.get('gravitationalConstant', None),
            'initial_position_radius': as_dict.get('initialPositionRadius', None),
            'initial_positions': as_dict.get('initialPositions', None),
            'integration': as_dict.get('integration', None),
            'link_length': as_dict.get('linkLength', None),
            'max_iterations': as_dict.get('maxIterations', None),
            'max_speed': as_dict.get('maxSpeed', None),
            'repulsive_force': as_dict.get('repulsiveForce', None),
            'theta': as_dict.get('theta', None),
            'type': as_dict.get('type', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'approximation': self.approximation,
            'attractiveForce': self.attractive_force,
            'enableSimulation': self.enable_simulation,
            'friction': self.friction,
            'gravitationalConstant': self.gravitational_constant,
            'initialPositionRadius': self.initial_position_radius,
            'initialPositions': self.initial_positions,
            'integration': self.integration,
            'linkLength': self.link_length,
            'maxIterations': self.max_iterations,
            'maxSpeed': self.max_speed,
            'repulsiveForce': self.repulsive_force,
            'theta': self.theta,
            'type': self.type
        }

        return untrimmed


class NetworkGraphOptions(GenericTypeOptions):
    """General options to apply to all Network Graph series types.

    A network graph is a type of relationship chart, where connnections (links)
    attract nodes (points) and other nodes repulse each other.

    .. figure:: ../../../_static/networkgraph-example.png
      :alt: NetworkGraph Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._color_index = None
        self._crisp = None
        self._draggable = None
        self._find_nearest_point_by = None
        self._layout_algorithm = None
        self._line_width = None
        self._link = None
        self._relative_x_value = None
        self._shadow = None
        self._zones = None

        self.color_index = kwargs.get('color_index', None)
        self.crisp = kwargs.get('crisp', None)
        self.draggable = kwargs.get('draggable', None)
        self.find_nearest_point_by = kwargs.get('find_nearest_point_by', None)
        self.layout_algorithm = kwargs.get('layout_algorithm', None)
        self.line_width = kwargs.get('line_width', None)
        self.link = kwargs.get('link', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)
        self.shadow = kwargs.get('shadow', None)
        self.zones = kwargs.get('zones', None)

        super().__init__(**kwargs)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def draggable(self) -> Optional[bool]:
        """If ``True``, indicates that the nodes are draggable. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable

    @draggable.setter
    def draggable(self, value):
        if value is None:
            self._draggable = None
        else:
            self._draggable = bool(value)

    @property
    def events(self) -> Optional[SimulationEvents]:
        """Event handlers for a network graph series.

        .. note::

          These event hooks can also be attached to the series at run time using the
          (JavaScript) ``Highcharts.addEvent()`` function.

        :rtype: :class:`SimulationEvents <highcharts_core.utility_classes.events.SimulationEvents>` or 
          :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(SimulationEvents)
    def events(self, value):
        self._events = value

    @property
    def find_nearest_point_by(self) -> Optional[str]:
        """Determines whether the series should look for the nearest point in both
        dimensions or just the x-dimension when hovering the series.

        If :obj:`None <python:None>`, defaults to ``'xy'`` for scatter series and ``'x'``
        for most other series. If the data has duplicate x-values, it is recommended to
        set this to ``'xy'`` to allow hovering over all points.

        Applies only to series types using nearest neighbor search (not direct hover) for
        tooltip.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._find_nearest_point_by

    @find_nearest_point_by.setter
    def find_nearest_point_by(self, value):
        self._find_nearest_point_by = validators.string(value, allow_empty = True)

    @property
    def layout_algorithm(self) -> Optional[LayoutAlgorithm]:
        """Configuration of how to lay out the Network Graph.

        :rtype: :class:`LayoutAlgorithm` or :obj:`None <python:None>`
        """
        return self._layout_algorithm

    @layout_algorithm.setter
    @class_sensitive(LayoutAlgorithm)
    def layout_algorithm(self, value):
        self._layout_algorithm = value

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the graph line. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def link(self) -> Optional[LinkOptions]:
        """Link style options.

        :rtype: :class:`LinkOptions` or :obj:`None <python:None>`
        """
        return self._link

    @link.setter
    @class_sensitive(LinkOptions)
    def link(self, value):
        self._link = value

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the tooltip. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or a boolean setting which hides the
          shadow or displays the default shadow.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if isinstance(value, bool):
            self._shadow = value
        elif not value:
            self._shadow = None
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def zones(self) -> Optional[List[Zone]]:
        """An array defining zones within a series. Defaults to :obj:`None <python:None>`.

        Zones can be applied to the X axis, Y axis or Z axis for bubbles, according to the
        :meth:`zone_axis <AreaOptions.zone_axis>` setting.

        .. warning::

          The zone definitions have to be in ascending order regarding to the value.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of
          :class:`Zone` instances
        """
        return self._zones

    @zones.setter
    @class_sensitive(Zone,
                     force_iterable = True)
    def zones(self, value):
        self._zones = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'draggable': as_dict.get('draggable', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'layout_algorithm': as_dict.get('layoutAlgorithm', None),
            'line_width': as_dict.get('lineWidth', None),
            'link': as_dict.get('link', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'zones': as_dict.get('zones', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'colorIndex': self.color_index,
            'crisp': self.crisp,
            'draggable': self.draggable,
            'findNearestPointBy': self.find_nearest_point_by,
            'layoutAlgorithm': self.layout_algorithm,
            'lineWidth': self.line_width,
            'link': self.link,
            'relativeXValue': self.relative_x_value,
            'shadow': self.shadow,
            'zones': self.zones
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.nodes import NodeOptions as cls
from highcharts_core.utility_classes.nodes import OrganizationNodeOptions as cls2
from highcharts_core.utility_classes.nodes import DependencyWheelNodeOptions as cls3

from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'color': '#cccccc',
      'color_index': 123,
      'data_labels': {
          'align': 'center',
          'allowOverlap': False,
          'backgroundColor': '#cccccc',
          'enabled': True,
          'overflow': 'justify',
          'shadow': False
      },
      'id': 'some-id-goes-here',
      'name': 'My Node Name',
      'offset_horizontal': 10,
      'offset_vertical': '5%'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_NodeOptions__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_NodeOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_NodeOptions_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_NodeOptions_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/nodes/01.js', False, None),

    ('utility_classes/nodes/error-01.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/nodes/01.js', True, None),

    ('utility_classes/nodes/error-01.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_NodeOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'color': '#cccccc',
      'color_index': 123,
      'column': 2,
      'data_labels': {
          'align': 'center',
          'allowOverlap': False,
          'backgroundColor': '#cccccc',
          'enabled': True,
          'overflow': 'justify',
          'shadow': False
      },
      'id': 'some-id-goes-here',
      'image': 'http://www.somewhere.com/',
      'layout': 'normal',
      'level': 2,
      'name': 'My Node Name',
      'offset_horizontal': 10,
      'offset_vertical': '5%',
      'title': 'Grand High Muckety-Muck'
    }, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_OrganizationNodeOptions__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_OrganizationNodeOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_OrganizationNodeOptions_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_OrganizationNodeOptions_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/nodes/02.js', False, None),

    ('utility_classes/nodes/error-03.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/nodes/02.js', True, None),

    ('utility_classes/nodes/error-03.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_OrganizationNodeOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)


STANDARD_PARAMS_3 = [
    ({}, None),
    ({
      'column': 2,
      'level': 2
    }, None),
    ({
      'column': 2,
      'level': 2,
      'color': '#cccccc',
      'color_index': 123,
      'data_labels': {
          'align': 'center',
          'allowOverlap': False,
          'backgroundColor': '#cccccc',
          'enabled': True,
          'overflow': 'justify',
          'shadow': False
      },
      'id': 'some-id-goes-here',
      'name': 'My Node Name',
      'offset_horizontal': 10,
      'offset_vertical': '5%'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_DependencyWheelNodeOptions__init__(kwargs, error):
    Class__init__(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_DependencyWheelNodeOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_DependencyWheelNodeOptions_from_dict(kwargs, error):
    Class_from_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_DependencyWheelNodeOptions_to_dict(kwargs, error):
    Class_to_dict(cls3, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/nodes/03.js', False, None),

    ('utility_classes/nodes/error-04.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/nodes/03.js', True, None),

    ('utility_classes/nodes/error-04.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_DependencyWheelNodeOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls3, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.exporting import Exporting as cls
from highcharts_core import errors, constants
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'accessibility': {
          'enabled': True
      },
      'allow_html': False,
      'buttons': {
          'context_button': constants.EnforcedNullType(),
          'test_key': {
              'text': 'Test Text',
              'enabled': True
          },
          'test_key2': {
              'text': 'Test Text 2',
              'enabled': True
          }
      },
      'enabled': True,
      'error': """function() { return true; }""",
      'fallback_to_export_server': True,
      'filename': 'test-filename',
      'form_attributes': {
          'test_attr': 'value-123',
          'test_attr2': 'value-345'
      },
      'lib_url': 'http://www.somewhere.com/',
      'menu_item_definitions': {
          'menu1': {
            'onclick': """function() { return true; }""",
            'text': 'My Menu Item',
            'textKey': 'my-menu-item',
            'separator': False
          },
          'menu2': {
            'onclick': """function() { return true; }""",
            'text': 'My Menu Item',
            'textKey': 'my-menu-item',
            'separator': False
          }
      },
      'pdf_font': {
          'bold': 'http://www.somefile.com/flie.otf',
          'bolditalic': 'http://www.somefile.com/flie.otf',
          'italic': 'http://www.somefile.com/flie.otf',
          'normal': 'http://www.somefile.com/flie.otf',
      },
      'print_max_width': 300,
      'scale': 1200,
      'show_table': True,
      'source_height': 120,
      'source_width': 300,
      'table_caption': 'Caption goes here',
      'type': 'image/png',
      'url': 'http://www.somewhere.com/',
      'use_multi_level_headers': False,
      'use_rowspan_headers': False,
      'width': 300
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('exporting/exporting/01.js', False, None),

    ('exporting/exporting/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('exporting/exporting/01.js', True, None),

    ('exporting/exporting/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.global_options.language.accessibility.series import SeriesTypeDescriptions as cls
from highcharts_core.global_options.language.accessibility.series import SeriesLanguageOptions as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'arearange': 'Area Range',
      'areasplinerange': 'Area Spline Range',
      'boxplot': 'Boxplot',
      'bubble': 'Bubble',
      'columnrange': 'Column Range',
      'errorbar': 'Errorbar',
      'funnel': 'Funnel',
      'pyramid': 'Pyramid',
      'waterfall': 'Waterfall'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_SeriesTypeDescriptions__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_SeriesTypeDescriptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_SeriesTypeDescriptions_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_SeriesTypeDescriptions_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('global_options/language/accessibility/series/01.js', False, None),

    ('global_options/language/accessibility/series/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('global_options/language/accessibility/series/01.js', True, None),

    ('global_options/language/accessibility/series/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_SeriesTypeDescriptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'description': 'Description goes here',
      'null_point_value': 'Null point value',
      'point_annotations_description': 'Point annotations description',
      'summary': {},
      'x_axis_description': 'X-Axis description',
      'y_axis_description': 'Y-Axis description'
    }, None),
    ({
      'description': 'Description goes here',
      'null_point_value': 'Null point value',
      'point_annotations_description': 'Point annotations description',
      'summary': {
          'bar': 'Bar',
          'barCombination': 'Bar combo',
          'boxplot': 'Boxplot',
          'boxplotCombination': 'Boxplot combo',
          'bubble': 'Bubble',
          'bubbleCombination': 'bubble combination',
          'column': 'column',
          'columnCombination': 'column_combination',
          'default': 'Default summary',
          'defaultCombination': 'Default combination',
          'line': 'Line',
          'lineCombination': 'Line combination',
          'map': 'Map',
          'mapCombination': 'Map combo',
          'mapbubble': 'Map Bubble',
          'mapbubbleCombination': 'Mapbubble Combo',
          'mapline': 'Map/Line',
          'maplineCombination': 'Map/Line Combo',
          'pie': 'Pie',
          'pieCombination': 'Pie Combo',
          'scatter': 'Scatter',
          'scatterCombination': 'Scatter Combo',
          'spline': 'Spline',
          'splineCombination': 'Spline Combination'
      },
      'x_axis_description': 'X-Axis description',
      'y_axis_description': 'Y-Axis description'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_SeriesLanguageOptions__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_SeriesLanguageOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_SeriesLanguageOptions_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_SeriesLanguageOptions_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('global_options/language/accessibility/series/02.js', False, None),

    ('global_options/language/accessibility/series/error-03.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('global_options/language/accessibility/series/02.js', True, None),

    ('global_options/language/accessibility/series/error-03.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_SeriesLanguageOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern

from highcharts_core.options.axes.numeric import NumericAxis
from highcharts_core.options.axes.crosshair import CrosshairOptions


class XAxis(NumericAxis):
    """Configuration settings for the X axis or category axis.

    Normally, this is the horizontal axis, though if the chart is inverted this becomes
    the vertical axis."""

    def __init__(self, **kwargs):
        self._crosshair = None
        self._height = None
        self._left = None
        self._line_color = None
        self._line_width = None
        self._minor_ticks_per_major = None
        self._show_empty = None
        self._top = None
        self._width = None

        self.crosshair = kwargs.get('crosshair', None)
        self.height = kwargs.get('height', None)
        self.left = kwargs.get('left', None)
        self.line_color = kwargs.get('line_color', None)
        self.line_width = kwargs.get('line_width', None)
        self.minor_ticks_per_major = kwargs.get('minor_ticks_per_major', None)
        self.show_empty = kwargs.get('show_empty', None)
        self.top = kwargs.get('top', None)
        self.width = kwargs.get('width', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'xAxis'

    @property
    def crosshair(self) -> Optional[CrosshairOptions]:
        """Configure a crosshair that follows either the mouse pointer or the hovered
        point. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CrosshairOptions` or :obj:`None <python:None>`
        """
        return self._crosshair

    @crosshair.setter
    def crosshair(self, value):
        if isinstance(value, bool):
            value = {
                'enabled': value
            }

        self._crosshair = validate_types(value, CrosshairOptions)

    @property
    def height(self) -> Optional[str | int | float | Decimal]:
        """The height of the axis when operating as the vertical axis, expressed either in
        pixels or as a percentage of the total plot height. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        if value is None:
            self._height = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except (TypeError, ValueError):
                value = validators.numeric(value, minimum = 0)

            self._height = value

    @property
    def left(self) -> Optional[str | int | float | Decimal]:
        """The left position of the axis when operating as a horizontal axis, expresesd
        as the pixel position relative to the chart or a percentage of the plot width,
        offset from the plot area's ``left`` position. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._left

    @left.setter
    def left(self, value):
        if value is None:
            self._left = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except (TypeError, ValueError):
                value = validators.numeric(value, minimum = 0)

            self._left = value

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the line marking the axis itself. Defaults to ``'#ccd6eb'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """The width of the marking the axis itself, in pixels. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def minor_ticks_per_major(self) -> Optional[int | float | Decimal]:
        """The number of minor ticks per major tick. Defaults to ``5``.
        
        .. note::
        
          Works for ``linear``, ``logarithmic`` and ``datetime`` axes.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._minor_ticks_per_major
        
    @minor_ticks_per_major.setter
    def minor_ticks_per_major(self, value):
        self._minor_ticks_per_major = validators.numeric(value,
                                                         allow_empty = True,
                                                         minimum = 0)
    @property
    def show_empty(self) -> Optional[bool]:
        """If ``True``, render the axis title and axis line even if the axis has no data.
        If ``False``, does not render the axis when the axis has no data. Defaults to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_empty

    @show_empty.setter
    def show_empty(self, value):
        if value is None:
            self._show_empty = None
        else:
            self._show_empty = bool(value)

    @property
    def top(self) -> Optional[str | int | float | Decimal]:
        """The top position of the axis when operating as a vertical axis, expresesd
        as the pixel position relative to the chart or a percentage of the plot width,
        offset from the plot area's ``top`` position. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._top

    @top.setter
    def top(self, value):
        if value is None:
            self._top = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except (TypeError, ValueError):
                value = validators.numeric(value, minimum = 0)

            self._top = value

    @property
    def width(self) -> Optional[str | int | float | Decimal]:
        """The width of the axis when operating as the vertical axis, expressed either in
        pixels or as a percentage of the total plot width. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._width = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except (TypeError, ValueError):
                value = validators.numeric(value, minimum = 0)

            self._width = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'align_ticks': as_dict.get('alignTicks', None),
            'allow_decimals': as_dict.get('allowDecimals', None),
            'alternate_grid_color': as_dict.get('alternateGridColor', None),
            'angle': as_dict.get('angle', None),
            'breaks': as_dict.get('breaks', None),
            'categories': as_dict.get('categories', None),
            'ceiling': as_dict.get('ceiling', None),
            'class_name': as_dict.get('className', None),
            'crossing': as_dict.get('crossing', None),
            'date_time_label_formats': as_dict.get('dateTimeLabelFormats', None),
            'end_on_tick': as_dict.get('endOnTick', None),
            'events': as_dict.get('events', None),
            'floor': as_dict.get('floor', None),
            'grid_line_color': as_dict.get('gridLineColor', None),
            'grid_line_dash_style': as_dict.get('gridLineDashStyle', None),
            'grid_line_interpolation': as_dict.get('gridLineInterpolation', None),
            'grid_line_width': as_dict.get('gridLineWidth', None),
            'grid_z_index': as_dict.get('gridZIndex', None),
            'id': as_dict.get('id', None),
            'labels': as_dict.get('labels', None),
            'linked_to': as_dict.get('linkedTo', None),
            'margin': as_dict.get('margin', None),
            'max': as_dict.get('max', None),
            'max_padding': as_dict.get('maxPadding', None),
            'min': as_dict.get('min', None),
            'minor_grid_line_color': as_dict.get('minorGridLineColor', None),
            'minor_grid_line_dash_style': as_dict.get('minorGridLineDashStyle', None),
            'minor_grid_line_width': as_dict.get('minorGridLineWidth', None),
            'minor_tick_color': as_dict.get('minorTickColor', None),
            'minor_tick_interval': as_dict.get('minorTickInterval', None),
            'minor_tick_length': as_dict.get('minorTickLength', None),
            'minor_tick_position': as_dict.get('minorTickPosition', None),
            'minor_ticks': as_dict.get('minorTicks', None),
            'minor_tick_width': as_dict.get('minorTickWidth', None),
            'min_padding': as_dict.get('minPadding', None),
            'min_range': as_dict.get('minRange', None),
            'min_tick_interval': as_dict.get('minTickInterval', None),
            'offset': as_dict.get('offset', None),
            'opposite': as_dict.get('opposite', None),
            'pane': as_dict.get('pane', None),
            'panning_enabled': as_dict.get('panningEnabled', None),
            'plot_bands': as_dict.get('plotBands', None),
            'plot_lines': as_dict.get('plotLines', None),
            'reversed': as_dict.get('reversed', None),
            'reversed_stacks': as_dict.get('reversedStacks', None),
            'show_first_label': as_dict.get('showFirstLabel', None),
            'show_last_label': as_dict.get('showLastLabel', None),
            'soft_max': as_dict.get('softMax', None),
            'soft_min': as_dict.get('softMin', None),
            'start_of_week': as_dict.get('startOfWeek', None),
            'start_on_tick': as_dict.get('startOnTick', None),
            'tick_amount': as_dict.get('tickAmount', None),
            'tick_color': as_dict.get('tickColor', None),
            'tick_interval': as_dict.get('tickInterval', None),
            'tick_length': as_dict.get('tickLength', None),
            'tickmark_placement': as_dict.get('tickmarkPlacement', None),
            'tick_pixel_interval': as_dict.get('tickPixelInterval', None),
            'tick_position': as_dict.get('tickPosition', None),
            'tick_positioner': as_dict.get('tickPositioner', None),
            'tick_positions': as_dict.get('tickPositions', None),
            'tick_width': as_dict.get('tickWidth', None),
            'title': as_dict.get('title', None),
            'type': as_dict.get('type', None),
            'unique_names': as_dict.get('uniqueNames', None),
            'units': as_dict.get('units', None),
            'visible': as_dict.get('visible', None),
            'z_index': as_dict.get('zIndex', None),
            'zoom_enabled': as_dict.get('zoomEnabled', None),

            'crosshair': as_dict.get('crosshair', None),
            'height': as_dict.get('height', None),
            'left': as_dict.get('left', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'minor_ticks_per_major': as_dict.get('minorTicksPerMajor', None),
            'show_empty': as_dict.get('showEmpty', None),
            'top': as_dict.get('top', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'crosshair': self.crosshair,
            'height': self.height,
            'left': self.left,
            'lineColor': self.line_color,
            'lineWidth': self.line_width,
            'minorTicksPerMajor': self.minor_ticks_per_major,
            'showEmpty': self.show_empty,
            'top': self.top,
            'width': self.width
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.events import MouseEvents

from highcharts_core.options.axes.labels import PlotBandLabel, PlotLineLabel


class PlotBand(HighchartsMeta):
    """Definition of a :term:`plot band`, which is a colored band stretching across the
    plot area which can be used to visually mark an interval along the axis.

    .. note::

      In a :term:`gauge chart`, a plot band on the :class:`YAxis` (value axis) will
      stretch along the perimeter of the gauge.

    """

    def __init__(self, **kwargs):
        self._border_color = None
        self._border_width = None
        self._class_name = None
        self._color = None
        self._events = None
        self._from_ = None
        self._id = None
        self._label = None
        self._to = None
        self._z_index = None

        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.events = kwargs.get('events', None)
        self.from_ = kwargs.get('from_', None)
        self.id = kwargs.get('id', None)
        self.label = kwargs.get('label', None)
        self.to = kwargs.get('to', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """Border color for the plot band. Defaults to :obj:`None <python:None>`.

        .. warning::

          Also requires :meth:`PlotBand.border_width` to be set.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """Border width for the plot band. Defaults to :obj:`None <python:None>`.

        .. warning::

          Also requires :meth:`PlotBand.border_color` to be set.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A custom class name (in addition to the default ``highcharts-plot-band``) to
        apply to each individual band. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the plot band. Defaults to ``#e6ebf5``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def events(self) -> Optional[MouseEvents]:
        """Event handlers for the Plot Band.

        :rtype: :class:`MouseEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(MouseEvents)
    def events(self, value):
        self._events = value

    @property
    def from_(self) -> Optional[int | float | Decimal]:
        """The point where the plot band starts, expressed in axis units. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._from

    @from_.setter
    def from_(self, value):
        self._from = validators.numeric(value, allow_empty = True)

    @property
    def id(self) -> Optional[str]:
        """An identifier used for identifying the plot band if calling (in JavaScript)
        ``Axis.removePlotBand()``. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def label(self) -> Optional[PlotBandLabel]:
        """Label options for the Plot Band. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`PlotBandLabel` or :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    @class_sensitive(PlotBandLabel)
    def label(self, value):
        self._label = value

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The end position of the plot band, expressed in axis units. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value, allow_empty = True)

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """The Z index for the plot band. Defaults to :obj:`None <python:None>`.

        .. warning::

          Using the same z-index as another element may give unpredictable results, as the
          last rendered element will be on top. Values from ``0`` to ``20`` make sense.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'events': as_dict.get('events', None),
            'from_': as_dict.get('from', None),
            'id': as_dict.get('id', None),
            'label': as_dict.get('label', None),
            'to': as_dict.get('to', None),
            'z_index': as_dict.get('zIndex', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'color': self.color,
            'events': self.events,
            'from': self.from_,
            'id': self.id,
            'label': self.label,
            'to': self.to,
            'zIndex': self.z_index
        }

        return untrimmed


class PlotLine(HighchartsMeta):
    """A line that stretches across the plot area, marking a specific value on an axis."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._color = None
        self._dash_style = None
        self._events = None
        self._id = None
        self._label = None
        self._value = None
        self._width = None
        self._z_index = None

        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.events = kwargs.get('events', None)
        self.id = kwargs.get('id', None)
        self.label = kwargs.get('label', None)
        self.value = kwargs.get('value', None)
        self.width = kwargs.get('width', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def class_name(self) -> Optional[str]:
        """A custom class name (in addition to the default ``highcharts-plot-line``) to
        apply to each individual line. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the line. Defaults to ``#999999``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the lines. Defaults to ``Solid``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        if not value:
            self._dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._dash_style = value

    @property
    def events(self) -> Optional[MouseEvents]:
        """Event handlers for the Plot Line.

        :rtype: :class:`MouseEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(MouseEvents)
    def events(self, value):
        self._events = value

    @property
    def label(self) -> Optional[PlotLineLabel]:
        """Label options for the Plot Line. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`PlotLineLabel` or :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    @class_sensitive(PlotLineLabel)
    def label(self, value):
        self._label = value

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The position of the line, expressed in axis units. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value):
        if hasattr(value, 'timestamp'):
            value = value.timestamp() * 1000

        self._value = validators.numeric(value, allow_empty = True)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width or thickness of the line, expressed in pixels. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, width):
        self._width = validators.numeric(width,
                                         allow_empty = True,
                                         minimum = 0)

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """The Z index for the plot line. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'dash_style': as_dict.get('dashStyle', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label': as_dict.get('label', None),
            'value': as_dict.get('value', None),
            'width': as_dict.get('width', None),
            'z_index': as_dict.get('zIndex', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'color': self.color,
            'dashStyle': self.dash_style,
            'events': self.events,
            'id': self.id,
            'label': self.label,
            'value': self.value,
            'width': self.width,
            'zIndex': self.z_index
        }

        return untrimmed



"""Defines a set of constants that are used throughout the library."""
import os
try:
    from dotenv import load_dotenv
    load_dotenv()
except ImportError:
    pass
try:
    import orjson as json
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        try:
            import simplejson as json
        except ImportError:
            import json

with open(os.path.join(os.path.dirname(__file__), 
                       './module_requirements.json'), 
          'r') as module_requirements:
    try:
        MODULE_REQUIREMENTS = json.load(module_requirements)
    except AttributeError:
        MODULE_REQUIREMENTS = json.loads(module_requirements.read())


class EnforcedNullType:
    def __eq__(self, other):
        return isinstance(other, self.__class__)


EnforcedNull = EnforcedNullType()


JAVASCRIPT_INDENT_SPACES = os.getenv('JAVASCRIPT_INDENT_SPACES') or 2
JAVASCRIPT_INDENT = ''
indent_count = 1
while indent_count < int(JAVASCRIPT_INDENT_SPACES):
    JAVASCRIPT_INDENT += ' '
    indent_count += 1

DEFAULT_COLORS = ["#7cb5ec", "#434348", "#90ed7d", "#f7a35c", "#8085e9", "#f15c80",
                  "#e4d354", "#2b908f", "#f45b5b", "#91e8e1"]

INCLUDE_LIBS = [
    'https://code.highcharts.com/highcharts.js',
    'https://code.highcharts.com/highcharts-more.js',
    'https://code.highcharts.com/highcharts-3d.js',
    'https://code.highcharts.com/modules/sonification.js',
    'https://code.highcharts.com/modules/accessibility.js',
    'https://code.highcharts.com/modules/annotations.js',
    'https://code.highcharts.com/modules/annotations-advanced.js',
    'https://code.highcharts.com/modules/sankey.js',
    'https://code.highcharts.com/modules/arc-diagram.js',
    'https://code.highcharts.com/modules/boost.js',
    'https://code.highcharts.com/modules/broken-axis.js',
    'https://code.highcharts.com/modules/bullet.js',
    'https://code.highcharts.com/modules/cylinder.js',
    'https://code.highcharts.com/modules/data.js',
    'https://code.highcharts.com/modules/datagrouping.js',
    'https://code.highcharts.com/modules/debugger.js',
    'https://code.highcharts.com/modules/dependency-wheel.js',
    'https://code.highcharts.com/modules/drag-panes.js',
    'https://code.highcharts.com/modules/draggable-points.js',
    'https://code.highcharts.com/modules/drilldown.js',
    'https://code.highcharts.com/modules/dumbbell.js',
    'https://code.highcharts.com/modules/export-data.js',
    'https://code.highcharts.com/modules/exporting.js',
    'https://code.highcharts.com/modules/funnel.js',
    'https://code.highcharts.com/modules/funnel3d.js',
    'https://code.highcharts.com/modules/heatmap.js',
    'https://code.highcharts.com/modules/item-series.js',
    'https://code.highcharts.com/modules/lollipop.js',
    'https://code.highcharts.com/modules/networkgraph.js',
    'https://code.highcharts.com/modules/no-data-to-display.js',
    'https://code.highcharts.com/modules/offline-exporting.js',
    'https://code.highcharts.com/modules/oldie.js',
    'https://code.highcharts.com/modules/organization.js',
    'https://code.highcharts.com/modules/parallel-coordinates.js',
    'https://code.highcharts.com/modules/pareto.js',
    'https://code.highcharts.com/modules/pictorial.js',
    'https://code.highcharts.com/modules/pyramid3d.js',
    'https://code.highcharts.com/modules/series-label.js',
    'https://code.highcharts.com/modules/series-on-point.js',
    'https://code.highcharts.com/modules/solid-gauge.js',
    'https://code.highcharts.com/modules/streamgraph.js',
    'https://code.highcharts.com/modules/sunburst.js',
    'https://code.highcharts.com/modules/tilemap.js',
    'https://code.highcharts.com/modules/timeline.js',
    'https://code.highcharts.com/modules/treegraph.js',
    'https://code.highcharts.com/modules/treemap.js',
    'https://code.highcharts.com/modules/variable-pie.js',
    'https://code.highcharts.com/modules/variwide.js',
    'https://code.highcharts.com/modules/vector.js',
    'https://code.highcharts.com/modules/venn.js',
    'https://code.highcharts.com/modules/windbarb.js',
    'https://code.highcharts.com/modules/wordcloud.js',
    'https://code.highcharts.com/modules/xrange.js',
]

INCLUDE_STR = """
    <script src="https://code.highcharts.com/highcharts.js"/>
    <script src="https://code.highcharts.com/highcharts-more.js"/>
    <script src="https://code.highcharts.com/highcharts-3d.js"/>
    <script src="https://code.highcharts.com/modules/sonification.js"/>
    <script src="https://code.highcharts.com/modules/accessibility.js"/>
    <script src="https://code.highcharts.com/modules/annotations.js"/>
    <script src="https://code.highcharts.com/modules/annotations-advanced.js"/>
    <script src="https://code.highcharts.com/modules/arc-diagram.js"/>
    <script src="https://code.highcharts.com/modules/bellcurve.js"/>
    <script src="https://code.highcharts.com/modules/boost.js"/>
    <script src="https://code.highcharts.com/modules/broken-axis.js"/>
    <script src="https://code.highcharts.com/modules/bullet.js"/>
    <script src="https://code.highcharts.com/modules/cylinder.js"/>
    <script src="https://code.highcharts.com/modules/data.js"/>
    <script src="https://code.highcharts.com/modules/datagrouping.js"/>
    <script src="https://code.highcharts.com/modules/debugger.js"/>
    <script src="https://code.highcharts.com/modules/dependency-wheel.js"/>
    <script src="https://code.highcharts.com/modules/drag-panes'
    <script src="https://code.highcharts.com/modules/draggable-points.js"/>
    <script src="https://code.highcharts.com/modules/drilldown.js"/>
    <script src="https://code.highcharts.com/modules/dumbbell.js"/>
    <script src="https://code.highcharts.com/modules/export-data.js"/>
    <script src="https://code.highcharts.com/modules/exporting.js"/>
    <script src="https://code.highcharts.com/modules/full-screen.js"/>
    <script src="https://code.highcharts.com/modules/funnel.js"/>
    <script src="https://code.highcharts.com/modules/funnel3d.js"/>
    <script src="https://code.highcharts.com/modules/histogram.js"/>
    <script src="https://code.highcharts.com/modules/item-series.js"/>
    <script src="https://code.highcharts.com/modules/lollipop.js"/>
    <script src="https://code.highcharts.com/modules/networkgraph.js"/>
    <script src="https://code.highcharts.com/modules/no-data-to-display.js"/>
    <script src="https://code.highcharts.com/modules/offline-exporting.js"/>
    <script src="https://code.highcharts.com/modules/oldie.js"/>
    <script src="https://code.highcharts.com/modules/organization.js"/>
    <script src="https://code.highcharts.com/modules/parallel-coordinates.js"/>
    <script src="https://code.highcharts.com/modules/pareto.js"/>
    <script src="https://code.highcharts.com/modules/pictorial.js"/>
    <script src="https://code.highcharts.com/modules/pyramid3d.js"/>
    <script src="https://code.highcharts.com/modules/sankey.js"/>
    <script src="https://code.highcharts.com/modules/series-label.js"/>
    <script src="https://code.highcharts.com/modules/series-on-point.js"/>
    <script src="https://code.highcharts.com/modules/solid-gauge.js"/>
    <script src="https://code.highcharts.com/modules/streamgraph.js"/>
    <script src="https://code.highcharts.com/modules/sunburst.js"/>
    <script src="https://code.highcharts.com/modules/tilemap.js"/>
    <script src="https://code.highcharts.com/modules/timeline.js"/>
    <script src="https://code.highcharts.com/modules/treegraph.js"/>
    <script src="https://code.highcharts.com/modules/treemap.js"/>
    <script src="https://code.highcharts.com/modules/variable-pie.js"/>
    <script src="https://code.highcharts.com/modules/variwide.js"/>
    <script src="https://code.highcharts.com/modules/vector.js"/>
    <script src="https://code.highcharts.com/modules/venn.js"/>
    <script src="https://code.highcharts.com/modules/windbarb.js"/>
    <script src="https://code.highcharts.com/modules/wordcloud.js"/>
    <script src="https://code.highcharts.com/modules/xrange.js"/>
"""

AXIS_TYPES = [
    'linear',
    'logarithmic',
    'datetime',
    'category'
]

SUPPORTED_CURSOR_VALUES = [
    'alias',
    'all-scroll',
    'auto',
    'cell',
    'col-resize',
    'context-menu',
    'copy',
    'crosshair',
    'default',
    'e-resize',
    'ew-resize',
    'grab',
    'grabbing',
    'help',
    'move',
    'n-resize',
    'ne-resize',
    'nesw-resize',
    'no-drop',
    'none',
    'not-allowed',
    'ns-resize',
    'nw-resize',
    'nwse-resize',
    'pointer',
    'progress',
    'row-resize',
    's-resize',
    'se-resize',
    'sw-resize',
    'text',
    'vertical-text',
    'w-resize',
    'wait',
    'zoom-in',
    'zoom-out'
]
SUPPORTED_DASH_STYLE_VALUES = [
    'Dash',
    'DashDot',
    'Dot',
    'LongDash',
    'LongDashDot',
    'LongDashDotDot',
    'ShortDash',
    'ShortDashDot',
    'ShortDashDotDot',
    'ShortDot',
    'Solid'
]

## ACCESSIBILITY DEFAULTS
DEFAULT_LANDMARK_VERBOSITY = 'all'
LANDMARK_VERBOSITY_VALUES = ['all',
                              'one',
                              'disabled'
                             ]

DEFAULT_LINKED_DESCRIPTION = '*[data-highcharts-chart="{index}"] + .highcharts-description'

DEFAULT_ACCESSIBILITY_POINT_VALUE_FORMAT = '{xDescription}{separator}{value}'
DEFAULT_AFTER_CHART_FORMAT = '{endOfChartMarker}'
DEFAULT_AXIS_RANGE_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
DEFAULT_BEFORE_CHART_FORMAT = ('<{headingTagName}>{chartTitle}</{headingTagName}><div>'
                               '{typeDescription}</div><div>{chartSubtitle}</div>'
                               '<div>{chartLongdesc}</div><div>{playAsSoundButton}</div>'
                               '<div>{viewTableButton}</div><div>{xAxisDescription}</div>'
                               '<div>{yAxisDescription}</div>'
                               '<div>{annotationsTitle}{annotationsList}</div>')
DEFAULT_DESCRIPTION_FORMAT = '{seriesDescription}{authorDescription}{axisDescription}'

## JAVASCRIPT LITERAL DEFAULTS

# TODO: DETERMINE IF THE FOLLOWING WILL ACTUALLY BE NEEDED
ALLOWED_JS_LITERAL_TYPES = [
    'FunctionDeclaration',
    'VariableDeclaration',
    'ClassDeclaration',
]


## ANNOTATIONS DEFAULTS
DEFAULT_DRAGGABLE = 'xy'
DEFAULT_LABEL_ALIGN = 'center'
DEFAULT_LABEL_BACKGROUND_COLOR = 'rgba(0, 0, 0, 0.75)'
DEFAULT_LABEL_BORDER_COLOR = '#000000'
DEFAULT_LABEL_BORDER_RADIUS = 3
DEFAULT_LABEL_BORDER_WIDTH = 1
DEFAULT_LABEL_CLASS_NAME = 'highcharts-no-tooltip'
DEFAULT_LABEL_OVERFLOW = 'justify'
DEFAULT_LABEL_FORMATTER = """function () { return defined(this.y) ? this.y : 'Annotation label'; }"""
DEFAULT_LABEL_PADDING = 5
DEFAULT_LABEL_SHAPE = 'callout'
DEFAULT_LABEL_VERTICAL_ALIGN = 'bottom'
DEFAULT_LABEL_X = 0
DEFAULT_LABEL_Y = -16


## SHAPES DEFAULTS
DEFAULT_SHAPES_FILL = 'rgba(0, 0, 0, 0.75)'
DEFAULT_SHAPES_R = 0
DEFAULT_SHAPES_SNAP = 2
DEFAULT_SHAPES_STROKE = 'rgba(0, 0, 0, 0.75)'
DEFAULT_SHAPES_STROKE_WIDTH = 1
DEFAULT_SHAPES_TYPE = 'rect'
SHAPES_ALLOWED_DASH_STYLES = ('Solid',
                              'ShortDash',
                              'ShortDot',
                              'ShortDashDot',
                              'ShortDashDotDot',
                              'Dot',
                              'Dash',
                              'LongDash',
                              'DashDot',
                              'LongDashDot',
                              'LongDashDotDot')

## BOOST DEFAULTS
DEFAULT_BOOST_PIXEL_RATIO = 1
DEFAULT_BOOST_SERIES_THRESHOLD = 50


## CAPTION DEFAULTS
DEFAULT_CAPTION_ALIGN = 'left'
DEFAULT_CAPTION_MARGIN = 15
DEFAULT_CAPTION_STYLE = '{"color": "#666666"}'
DEFAULT_CAPTION_VERTICAL_ALIGN = 'bottom'
DEFAULT_CAPTION_X = 0
DEFAULT_CAPTION_Y = None


## CHART DEFAULTS
DEFAULT_CHART_BACKGROUND_COLOR = '#ffffff'
DEFAULT_CHART_BORDER_COLOR = '#335cad'
DEFAULT_CHART_BORDER_RADIUS = 0
DEFAULT_CHART_BORDER_WIDTH = 0
DEFAULT_CHART_COLOR_COUNT = 10
DEFAULT_CHART_SERIES_TYPE = 'line'
DEFAULT_CHART_PLOT_BACKGROUND_COLOR = None
DEFAULT_CHART_PLOT_BORDER_COLOR = '#cccccc'
DEFAULT_CHART_PLOT_BORDER_WIDTH = 0
DEFAULT_CHART_SELECTION_MARKER_FILL = 'rgba(51,92,173,0.25)'
DEFAULT_CHART_SPACING_BOTTOM = 15
DEFAULT_CHART_SPACING_LEFT = 10
DEFAULT_CHART_SPACING_TOP = 10
DEFAULT_CHART_SPACING_RIGHT = 10
DEFAULT_CHART_SPACING = [DEFAULT_CHART_SPACING_TOP,
                         DEFAULT_CHART_SPACING_RIGHT,
                         DEFAULT_CHART_SPACING_BOTTOM,
                         DEFAULT_CHART_SPACING_LEFT]
DEFAULT_CHART_STYLE = '{"fontFamily": "\"Lucida Grande\", \"Lucida Sans Unicode\", Verdana, Arial, Helvetica, sans-serif","fontSize":"12px"}'
DEFAULT_CHART_TYPE = 'line'
DEFAULT_CHART_PARALLEL_AXES = {
    'line_width': 1,
    'gridlines_width': 0,
    'title': {
        'text': '',
        'reserve_space': False
    },
    'labels': {
        'x': 0,
        'y': 0,
        'align': 'center',
        'reserve_space': False
    },
    'offset': 0
}


## RESET ZOOM BUTTON DEFAULTS
DEFAULT_RESET_ZOOM_BUTTON_POSITION = {
    'align': 'right',
    'vertical_align': 'top',
    'x': -10,
    'y': 10
}
DEFAULT_RESET_ZOOM_BUTTON_THEME = {
    'zIndex': 6
}


## CREDITS DEFAULTS
DEFAULT_CREDITS_HREF = 'https://www.highcharts.com?credits'
DEFAULT_CREDITS_STYLE = {
    'color': '#999999',
    'cursor': 'pointer',
    'fontSize': '9px'
}
DEFAULT_CREDITS_TEXT = 'Highcharts'


## DRILLDOWN DEFAULTS
DEFAULT_DRILLDOWN_ACTIVE_AXIS_LABEL_STYLE = {
    "cursor": "pointer",
    "color": "#003399",
    "fontWeight": "bold",
    "textDecoration": "underline"
}
DEFAULT_DRILLDOWN_ACTIVE_DATA_LABEL_STYLE = {
    'color': '#003399',
    'cursor': 'pointer',
    'fontWeight': 'bold',
    'textDecoration': 'underline'
}


## DRILLDOWN BREAADCRUMB DEFAULTS
DEFAULT_BREADCRUMBS_POSITION = {
    'align': 'left',
    'vertical_align': 'top',
    'x': 0,
    'y': None
}
DEFAULT_BREADCRUMBS_RELATIVE_TO = 'plotBox'
DEFAULT_BREADCRUMBS_STYLE = None
DEFAULT_BREADCRUMBS_SEPARATOR_STYLE = {
    'color': '#666666'
}
DEFAULT_BREADCRUMBS_SEPARATOR_TEXT = '/'


## EXPORTING DEFAULTS
DEFAULT_EXPORTING_FILENAME = 'chart'
DEFAULT_EXPORTING_LIB_URL = None
DEFAULT_EXPORTING_MENU_ITEM_DEFINITIONS = {
    "viewFullscreen": {},
    "printChart": {},
    "separator": {},
    "downloadPNG": {},
    "downloadJPEG": {},
    "downloadPDF": {},
    "downloadSVG": {}
}
DEFAULT_EXPORTING_PRINT_MAX_WIDTH = 780
DEFAULT_EXPORTING_SCALE = 2
DEFAULT_EXPORTING_TYPE = 'image/png'
DEFAULT_EXPORTING_URL = 'https://export.highcharts.com/'

## EXPORTING: CONTEXT BUTTON
DEFAULT_CONTEXT_BUTTON_CLASS_NAME = 'highcharts-contextbutton'
DEFAULT_CONTEXT_MENU_CLASS_NAME = 'highcharts-contextmenu'
DEFAULT_CONTEXT_MENU_ITEMS = [
    "viewFullscreen",
    "printChart",
    "separator",
    "downloadPNG",
    "downloadJPEG",
    "downloadPDF",
    "downloadSVG"
]

## EXPORTING: CSV
DEFAULT_CSV_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
DEFAULT_CSV_LINE_DELIMITER = '\n'


## LANG
DEFAULT_LANG_CONTEXT_BUTTON_TITLE = 'Chart context menu'
DEFAULT_LANG_DOWNLOAD_CSV = 'Download CSV'
DEFAULT_LANG_DOWNLOAD_JPEG = 'Download JPEG'
DEFAULT_LANG_DOWNLOAD_PDF = 'Download PDF'
DEFAULT_LANG_DOWNLOAD_PNG = 'Download PNG'
DEFAULT_LANG_DOWNLOAD_SVG = 'Download SVG'
DEFAULT_LANG_DOWNLOAD_XLS = 'Download Excel'
DEFAULT_LANG_DRILLUP_TEXT = None
DEFAULT_LANG_EXIT_FULLSCREEN = 'Exit from full screen'
DEFAULT_LANG_HIDE_DATA = 'Hide data table'
DEFAULT_LANG_INVALID_DATE = ''
DEFAULT_LANG_LOADING = 'Loading...'
DEFAULT_LANG_MAIN_BREADCRUM = 'Main'
DEFAULT_LANG_MONTHS = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December"
]
DEFAULT_LANG_NO_DATA = 'No data to display'
DEFAULT_LANG_NUMERIC_SYMBOL_MAGNITUDE = 1000
DEFAULT_LANG_NUMERIC_SYMBOLS = ["k", "M", "G", "T", "P", "E"]
DEFAULT_LANG_PRINT_CHART = 'Print chart'
DEFAULT_LANG_RESET_ZOOM = 'Reset zoom'
DEFAULT_LANG_RESET_ZOOM_TITLE = 'Reset zoom level 1:1'
DEFAULT_LANG_SHORT_MONTHS = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec"
]
DEFAULT_LANG_SHORT_WEEKDAYS = None
DEFAULT_LANG_THOUSANDS_SEP = '\u0020'
DEFAULT_LANG_VIEW_DATA = 'View data table'
DEFAULT_LANG_VIEW_FULLSCREEN = 'View in full screen'
DEFAULT_LANG_WEEKDAYS = [
    "Sunday",
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday"
]

## LANGUAGE: ACCESSIBILITY
DEFAULT_LANG_ACCESSIBILITY_CHART_CONTAINER_LABEL = '{title}. Highcharts interactive chart.'
DEFAULT_LANG_ACCESSIBILITY_CREDITS = 'Chart credits: {creditsStr}'
DEFAULT_LANG_ACCESSIBILITY_DEFAULT_CHART_TITLE = 'Chart'
DEFAULT_LANG_ACCESSIBILITY_DRILLUP_BUTTON = '{buttonText}'
DEFAULT_LANG_ACCESSIBILITY_GRAPHIC_CONTAINER_LABEL = ''
DEFAULT_LANG_ACCESSIBILITY_SVG_CONTAINER_LABEL = 'Interactive chart'
DEFAULT_LANG_ACCESSIBILITY_SVG_CONTAINER_TITLE = ''
DEFAULT_LANG_ACCESSIBILITY_THOUSANDSSEP = ','

DEFAULT_LANG_ACS_ANNOUNCE_NEW_DATA = 'Updated data for chart {chartTitle}'
DEFAULT_LANG_ACS_ANNOUNCE_NEW_POINT_MULTIPLE = 'New data point in chart {chartTitle}: {pointDesc}'
DEFAULT_LANG_ACS_ANNOUNCE_NEW_POINT_SINGLE = 'New data point: {pointDesc}'
DEFAULT_LANG_ACS_ANNOUNCE_NEW_SERIES_MULTIPLE = 'New data series in chart {chartTitle}: {seriesDesc}'
DEFAULT_LANG_ACS_ANNOUNCE_NEW_SERIES_SINGLE = 'New data series: {seriesDesc}'

DEFAULT_LANG_ACS_AXIS_RANGE_CATEGORIES = 'Data range: {numCategories} categories.'
DEFAULT_LANG_ACS_AXIS_RANGE_FROM_TO = 'Data ranges from {rangeFrom} to {rangeTo}.'
DEFAULT_LANG_ACS_AXIS_TIME_RANGE_DAYS = 'Data range: {range} days.'
DEFAULT_LANG_ACS_AXIS_TIME_RANGE_HOURS = 'Data range: {range} hours.'
DEFAULT_LANG_ACS_AXIS_TIME_RANGE_MINUTES = 'Data range: {range} minutes.'
DEFAULT_LANG_ACS_AXIS_TIME_RANGE_SECONDS = 'Data range: {range} seconds.'
DEFAULT_LANG_ACS_X_AXIS_DESCRIPTION_PLURAL = 'The chart has {numAxes} X axes displaying {#each(names, -1), }and {names[-1]}.'
DEFAULT_LANG_ACS_X_AXIS_DESCRIPTION_SINGULAR = 'The chart has 1 X axis displaying {names[0]}. {ranges[0]}'
DEFAULT_LANG_ACS_Y_AXIS_DESCRIPTION_PLURAL = 'The chart has {numAxes} Y axes displaying {#each(names, -1), }and {names[-1]}.'
DEFAULT_LANG_ACS_Y_AXIS_DESCRIPTION_SINGULAR = 'The chart has 1 Y axis displaying {names[0]}. {ranges[0]}'

DEFAULT_LANG_ACS_CHART_TYPES_BAR_MULTIPLE = 'Bar chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_BAR_SINGLE = 'Bar chart with {numPoints} {#plural(numPoints, bars, bar)}.'
DEFAULT_LANG_ACS_CHART_TYPES_BOXPLOT_MULTIPLE = 'Boxplot with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_BOXPLOT_SINGLE = 'Boxplot with {numPoints} {#plural(numPoints, boxes, box)}.'
DEFAULT_LANG_ACS_CHART_TYPES_BUBBLE_MULTIPLE = 'Bubble chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_BUBBLE_SINGLE = 'Bubble chart with {numPoints} {#plural(numPoints, bubbles, bubble)}.'
DEFAULT_LANG_ACS_CHART_TYPES_COLUMN_MULTIPLE = 'Bar chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_COLUMN_SINGLE = 'Bar chart with {numPoints} {#plural(numPoints, bars, bar)}.'
DEFAULT_LANG_ACS_CHART_TYPES_COMBO = 'Combination chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_DEFAULT_MULTIPLE = 'Chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_DEFAULT_SINGLE = 'Chart with {numPoints} data {#plural(numPoints, points, point)}.'
DEFAULT_LANG_ACS_CHART_TYPES_EMPTY_CHART = 'Empty chart'
DEFAULT_LANG_ACS_CHART_TYPES_LINE_MULTIPLE = 'Line chart with {numSeries} lines.'
DEFAULT_LANG_ACS_CHART_TYPES_LINE_SINGLE = 'Line chart with {numPoints} data {#plural(numPoints, points, point)}.'
DEFAULT_LANG_ACS_CHART_TYPES_MAP_TYPE_DESCRIPTION = 'Map of {mapTitle} with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_PIE_MULTIPLE = 'Pie chart with {numSeries} pies.'
DEFAULT_LANG_ACS_CHART_TYPES_PIE_SINGLE = 'Pie chart with {numPoints} {#plural(numPoints, slices, slice)}.'
DEFAULT_LANG_ACS_CHART_TYPES_SCATTER_MULTIPLE = 'Scatter chart with {numSeries} data series.'
DEFAULT_LANG_ACS_CHART_TYPES_SCATTER_SINGLE = 'Scatter chart with {numPoints} {#plural(numPoints, points, point)}.'
DEFAULT_LANG_ACS_CHART_TYPES_SPLINE_MULTIPLE = 'Line chart with {numSeries} lines.'
DEFAULT_LANG_ACS_CHART_TYPES_SPLINE_SINGLE = 'Line chart with {numPoints} data {#plural(numPoints, points, point)}.'
DEFAULT_LANG_ACS_CHART_TYPES_UNKNOWN_MAP = 'Map of unspecified region with {numSeries} data series.'

DEFAULT_LANG_ACS_EXPORTING_CHART_MENU_LABEL = 'Chart menu'
DEFAULT_LANG_ACS_EXPORTING_MENU_BTN_LABEL = 'View chart menu, {chartTitle}'


DEFAULT_LANG_ACS_LEGEND_ITEM = 'Show {itemName}'
DEFAULT_LANG_ACS_LEGEND_LABEL = 'Chart legend: {legendTitle}'
DEFAULT_LANG_ACS_LEGEND_LABEL_NO_TITLE = 'Toggle series visibility, {chartTitle}'


DEFAULT_LANG_ACS_RANGE_SELECTOR_CLICK_BTN_ANNOUNCEMENT = 'Viewing {axisRangeDescription}'
DEFAULT_LANG_ACS_RANGE_SELECTOR_DROPDOWN_LBL = '{rangeTitle}'
DEFAULT_LANG_ACS_RANGE_SELECTOR_MAX_INPUT_LBL = 'Select end date.'
DEFAULT_LANG_ACS_RANGE_SELECTOR_MIN_INPUT_LBL = 'Select start date.'


DEFAULT_LANG_ACS_SRS_AFTER_REGION_LBL = ''
DEFAULT_LANG_ACS_SRS_BEFORE_REGION_LBL = ''
DEFAULT_LANG_ACS_SRS_END_OF_CHART_MRKR = 'End of interactive chart.'


DEFAULT_LANG_ACS_SRS_ANNOTATION_DESCRIPTION_MULTIPLE_PTS = '{annotationText}. Related to {annotationPoint}{ Also related to, #each(additionalAnnotationPoints)}'
DEFAULT_LANG_ACS_SRS_ANNOTATION_DESCRIPTION_NO_PTS = '{annotationText}'
DEFAULT_LANG_ACS_SRS_ANNOTATION_DESCRIPTION_SINGLE_PT = '{annotationText}. Related to {annotationPoint}'
DEFAULT_LANG_ACS_SRS_ANNOTATION_HEADING = 'Chart annotations summary'


DEFAULT_LANG_ACS_SERIES_DESCRIPTION = ''
DEFAULT_LANG_ACS_SERIES_NULL_PT_VALUE = ''
DEFAULT_LANG_ACS_SERIES_PT_ANNOTATIONS_DESCRIPTION = ''
DEFAULT_LANG_ACS_SERIES_XAXIS_DESCRIPTION = ''
DEFAULT_LANG_ACS_SERIES_YAXIS_DESCRIPTION = ''


DEFAULT_LANG_ACS_SERIES_SUMMARY = {
    'bar': '{series.name}, bar series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bars, bar)}.',
    'bar_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Bar series with {series.points.length} {#plural(series.points.length, bars, bar)}.',
    'boxplot': '{series.name}, boxplot {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, boxes, box)}.',
    'boxplot_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Boxplot with {series.points.length} {#plural(series.points.length, boxes, box)}.',
    'bubble': '{series.name}, bubble series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.',
    'bubble_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Bubble series with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.',
    'column': '{series.name}, bar series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bars, bar)}.',
    'column_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Bar series with {series.points.length} {#plural(series.points.length, bars, bar)}.',
    'default': '{series.name}, series {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'default_combination': '{series.name}, series {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'line': '{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'line_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'map': '{series.name}, map {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, areas, area)}.',
    'map_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Map with {series.points.length} {#plural(series.points.length, areas, area)}.',
    'mapbubble': '{series.name}, bubble series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.',
    'mapbubble_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Bubble series with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.',
    'mapline': '{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'mapline_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'pie': '{series.name}, pie {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, slices, slice)}.',
    'pie_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Pie with {series.points.length} {#plural(series.points.length, slices, slice)}.',
    'scatter': '{series.name}, scatter plot {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, points, point)}.',
    'scatter_combination': '{series.name}, series {seriesNumber} of {chart.series.length}, scatter plot with {series.points.length} {#plural(series.points.length, points, point)}.',
    'spline': '{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.',
    'spline_combination': '{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.',
}


DEFAULT_LANG_ACS_SERIES_TYPES = {
    'arearange': 'Arearange charts are line charts displaying a range between a lower and higher value for each point.',
    'areasplinerange': 'These charts are line charts displaying a range between a lower and higher value for each point.',
    'boxplot': 'Box plot charts are typically used to display groups of statistical data. Each data point in the chart can have up to 5 values: minimum, lower quartile, median, upper quartile, and maximum.',
    'bubble': 'Bubble charts are scatter charts where each data point also has a size value.',
    'columnrange': 'Columnrange charts are column charts displaying a range between a lower and higher value for each point.',
    'errorbar': 'Errorbar series are used to display the variability of the data.',
    'funnel': 'Funnel charts are used to display reduction of data in stages.',
    'pyramid': 'Pyramid charts consist of a single pyramid with item heights corresponding to each point value.',
    'waterfall': 'A waterfall chart is a column chart where each column contributes towards a total end value.'
}


## LANGUAGE: ACCESSIBILITY: SONIFICATION
DEFAULT_LANG_ACS_SONIFICATION_PLAY_AS_SOUND_BTN_TXT = 'Play as sound, {chartTitle}'
DEFAULT_LANG_ACS_SONIFICATION_PLAY_AS_SOUND_CLK_ANNOUNCEMENT = 'Play'


## LANGUAGES: ACCESSIBILITY: TABLE
DEFAULT_LANG_ACS_TABLE_SUMMARY = 'Table representation of chart.'
DEFAULT_LANG_ACS_TABLE_VIEW_AS_DATA_TABLE = 'View as data table, {chartTitle}'


## LANGUAGES: ACCESSIBILITY: ZOOM
DEFAULT_LANG_ACS_ZOOM_MAP_ZOOM_IN = 'Zoom chart'
DEFAULT_LANG_ACS_ZOOM_MAP_ZOOM_OUT = 'Zoom out chart'
DEFAULT_LANG_ACS_ZOOM_RESET_ZOOM_BTN = 'Reset zoom'


## LANGUAGES: EXPORT DATA
DEFAULT_LANG_EXPORT_DATA = {
    'annotation_header': 'Annotations',
    'category_datetime_header': 'DateTime',
    'category_header': 'Category'
}


## LANGUAGES: NAVIGATION
DEFAULT_LANG_NAVIGATION = {
    'add_button': 'add',
    'algorithm': 'Algorithm',
    'arrow_infinity_line': 'Arrow line',
    'arrow_ray': 'Arrow ray',
    'arrow_segment': 'Arrow segment',
    'average': 'Average',
    'background': 'Background',
    'background_color': 'Background color',
    'background_colors': 'Background colors',
    'border_color': 'Border color',
    'border_radius': 'Border radius',
    'border_width': 'Border width',
    'bottom_band': 'Bottom band',
    'circle': 'Circle',
    'clear_filter': 'X Clear Filter',
    'color': 'Color',
    'connector': 'Connector',
    'crooked3': 'Crooked 3 line',
    'crooked5': 'Crooked 5 line',
    'crosshairX': 'Crosshair X',
    'crosshairY': 'Crosshair Y',
    'decimals': 'Decimals',
    'deviation': 'Deviation',
    'edit_button': 'edit',
    'elliott3': 'Elliott 3 line',
    'elliott5': 'Elliott 5 line',
    'ellipse': 'Ellipse',
    'factor': 'Factor',
    'fast_avg_period': 'Fast average period',
    'fibonacci': 'Fibonacci',
    'fibonacci_time_zones': 'Fibonacci Time Zones',
    'fill': 'Fill',
    'flags': 'Flags',
    'font_size': 'Font size',
    'format': 'Text',
    'height': 'Height',
    'high_index': 'High index',
    'horizontal_line': 'Horizontal line',
    'increment': 'Increment',
    'index': 'Index',
    'infinity_line': 'Infinity line',
    'initial_acceleration_factor': 'Initial acceleration factor',
    'inner_background': 'Inner background',
    'label': 'Label',
    'label_options': 'Label options',
    'labels': 'Labels',
    'line': 'Line',
    'lines': 'Lines',
    'long_period': 'Long period',
    'low_index': 'Low index',
    'max_acceleration_factor': 'Max acceleration factor',
    'measure': 'Measure',
    'measure_x': 'Measure X',
    'measure_xy': 'Measure XY',
    'measure_y': 'Measure Y',
    'multiplier': 'Multiplier',
    'multiplier_atr': 'ATR multiplier',
    'name': 'Name',
    'no_filter_match': 'No match',
    'outer_background': 'Outer background',
    'padding': 'Padding',
    'parallel_channel': 'Parallel channel',
    'period': 'Period',
    'period_atr': 'ATR period',
    'periods': 'Periods',
    'period_senkou_span_b': 'Senkou Span B period',
    'period_tenkan': 'Tenkan period',
    'pitchfork': 'Pitchfork',
    'ranges': 'Ranges',
    'ray': 'Ray',
    'rectangle': 'Rectangle',
    'remove_button': 'remove',
    'save_button': 'save',
    'search_indicators': 'Search Indicators',
    'segment': 'Segment',
    'series': 'Series',
    'shape_options': 'Shape options',
    'shapes': 'Shape options',
    'short_period': 'Short period',
    'signal_period': 'Signal period',
    'simple_shapes': 'Simple shapes',
    'slow_avg_period': 'Slow average period',
    'standard_deviation': 'Standard deviation',
    'stroke': 'Line color',
    'stroke_width': 'Line width',
    'style': 'Style',
    'time_cycles': 'Time Cycles',
    'title': 'Title',
    'top_band': 'Top band',
    'tunnel': 'Tunnel',
    'type_options': 'Details',
    'vertical_arrow': 'Vertical arrow',
    'vertical_counter': 'Vertical counter',
    'vertical_label': 'Vertical label',
    'vertical_line': 'Vertical line',
    'volume': 'Volume',
    'x_axis_unit': 'x-axis unit'
}


## DEFAULTS: LEGEND
DEFAULT_LEGEND = {
    'accessibility': None,
    'align': 'center',
    'align_columns': True,
    'background_color': None,
    'border_color': '#999999',
    'border_radius': 0,
    'border_width': 0,
    'bubble_legend': None,
    'class_name': 'highcharts-no-tooltip',
    'enabled': False,
    'floating': False,
    'item_checkbox_style': '{"width": "13px", "height": "13px", "position":"absolute"}',
    'item_distance': 20,
    'item_hidden_style': '{"color": "#cccccc"}',
    'item_hover_style': '{"color": "#000000"}',
    'item_margin_bottom': 0,
    'item_margin_top': 0,
    'item_style': '{"color": "#333333", "cursor": "pointer", "fontSize": "12px", "fontWeight": "bold", "textOverflow": "ellipsis"}',
    'item_width': None,
    'label_format': '{name}',
    'label_formatter': None,
    'layout': 'horizontal',
    'margin': 12,
    'max_height': None,
    'navigation': {
        'active_color': '#003399',
        'arrow_size': 12,
        'inactive_color': '#cccccc',
        'style': None
    },
    'padding': 8,
    'reversed': False,
    'rtl': False,
    'shadow': False,
    'square_symbol': True,
    'symbol_height': None,
    'symbol_padding': 5,
    'symbol_radius': None,
    'symbol_width': None,
    'title': {
        'style': '{"fontWeight": "bold"}'
    },
    'use_html': False,
    'vertical_align': 'bottom',
    'width': None,
    'x': 0,
    'y': 0
}
DEFAULT_BUBBLE_LEGEND = {
    'border_color': None,
    'border_width': 2,
    'class_name': None,
    'color': None,
    'connector_class_name': None,
    'connector_color': None,
    'connector_distance': None,
    'connector_width': None,
    'enabled': False,
    'labels': {
        'align': 'right',
        'allow_overlap': False,
        'class_name': None,
        'format': '',
        'formatter': None,
        'style': None,
        'x': 0,
        'y': 0
    },
    'legend_index': 0,
    'max_size': 60,
    'min_size': 10,
    'ranges': None,
    'size_by': 'area',
    'size_by_absolute_value': False,
    'z_index': 1,
    'z_threshold': 0

}


## LOADING
DEFAULT_LOADING = {
    'hide_duration': 100,
    'label_style': '{"fontWeight": "bold", "position": "relative", "top": "45%"}',
    'show_duration': 100,
    'style': '{"position": "absolute", "backgroundColor": "#ffffff", "opacity": 0.5, "textAlign": "center"}'
}


## NAVIGATION
DEFAULT_NAVIGATION = {
    'bindings_class_name': 'highcharts-bindings-container',
    'icons_url': 'https://code.highcharts.com/@product.version@/gfx/stock-icons/'
}

## PANE
DEFAULT_PANE = {
    'center': [['50%', '50%']],
    'end_angle': None,
    'inner_size': '0%',
    'size': '85%',
    'start_angle': 0
}
DEFAULT_PANE_BACKGROUND = {
    'background_color': '{ "linearGradient": { "x1": 0, "y1": 0, "x2": 0, "y2": 1 }, stops: [[0, "#ffffff"], [1, "#e6e6e6"]] }',
    'border_color': '#cccccc',
    'border_width': 1,
    'class_name': 'highcharts-pane',
    'inner_radius': 0,
    'outer_radius': '105%',
    'shape': 'circle'
}


## TOOLTIP
DEFAULT_TOOLTIP = {
    'border_radius': 3,
    'border_width': 1,
    'distance': 16,
    'footer_format': '',
    'header_shape': 'callout',
    'hide_delay': 500,
    'padding': 8,
    'shape': 'callout',
    'snap': None
}


DEFAULT_DATE_TIME_LABEL_FORMATS = {
    'day': '%A, %b %e, %Y',
    'hour': '%A, %b %e, %H:%M',
    'millisecond': '%A, %b %e, %H:%M:%S.%L',
    'minute': '%A, %b %e, %H:%M',
    'month': '%B %Y',
    'second': '%A, %b %e, %H:%M:%S',
    'week': 'Week from %A, %b %e, %Y',
    'year': '%Y'
}


## DATA LABELS
DEFAULT_DATA_LABEL = {
    'border_radius': 0,
    'border_width': 0,
    'class_name': None,
    'color': None,
    'crop': True,
    'defer': True,
    'format': 'point.value',
    'overflow': 'justify',
    'padding': 5,
    'position': 'center',
    'rotation': 0,
    'shadow': False,
    'shape': 'square',
    'style': None,
    'vertical_align': None,
    'x': 0,
    'y': None,
    'z': 6
}


## MARKER
DEFAULT_MARKER = {
    'enabled_threshold': 2,
    'line_color': '#ffffff',
    'line_width': 0,
    'radius': 4,
    'symbol': None,
    'width': None
}


## DAYS OF WEEK
DAYS_OF_WEEK = {
    'sunday': 0,
    'monday': 1,
    'tuesday': 2,
    'wednesday': 3,
    'thursday': 4,
    'friday': 5,
    'saturday': 6
}

## SONIFICATION INSTRUMENTS
INSTRUMENT_PRESETS = [
    'piano',
    'flute',
    'saxophone',
    'trumpet',
    'sawsynth',
    'wobble',
    'basic1',
    'basic2',
    'sine',
    'sineGlide',
    'triangle',
    'square',
    'sawtooth',
    'noise',
    'filteredNoise',
    'wind',
]

EXPORT_SERVER_UNSUPPORTED_SERIES_TYPES = [
    'pictorial',
    'flowmap',
    'geoheatmap',
    'treegraph',
]


from highcharts_core.options.plot_options.funnel import FunnelOptions, Funnel3DOptions


class PyramidOptions(FunnelOptions):
    """General options to apply to all Pyramid series types.

    A pyramid series is a special type of funnel, without neck and reversed by
    default.

    .. figure:: ../../../_static/pyramid-example.png
      :alt: Pyramid Example Chart
      :align: center

    """
    pass

class Pyramid3DOptions(Funnel3DOptions):
    """General options to apply to all Pyramid 3D series types.

    A pyramid 3d series is a special type of funnel, without neck and reversed by
    default.

    .. figure:: ../../../_static/pyramid_3d-example.png
      :alt: Pyramid 3D Example Chart
      :align: center

    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
